@@ -17,6 +17,7 @@
 #include "access/htup_details.h"
 #include "access/sysattr.h"
 #include "access/transam.h"
+#include "catalog/pg_aggregate.h"
 #include "catalog/pg_collation.h"
 #include "catalog/pg_namespace.h"
 #include "catalog/pg_operator.h"
@@ -41,6 +42,7 @@
 #include "utils/lsyscache.h"
 #include "utils/syscache.h"
 #include "utils/timestamp.h"
+#include "utils/typcache.h"
 
 
 static char *mysql_quote_identifier(const char *str, char quotechar);
@@ -56,9 +58,12 @@ typedef struct foreign_glob_cxt
 	/*
 	 * For join pushdown, only a limited set of operators are allowed to be
 	 * pushed.  This flag helps us identify if we are walking through the list
-	 * of join conditions.
+	 * of join conditions. Also true for aggregate relations to restrict
+	 * aggregates for specified list.
 	 */
-	bool		is_join_cond;	/* true for join relations */
+	bool		is_remote_cond;	/* true for join or aggregate relations */
+	Relids		relids;			/* relids of base relations in the underlying
+								 * scan */
 } foreign_glob_cxt;
 
 /*
@@ -85,6 +90,9 @@ typedef struct deparse_expr_cxt
 {
 	PlannerInfo *root;			/* global planner state */
 	RelOptInfo *foreignrel;		/* the foreign relation we are planning for */
+	RelOptInfo *scanrel;		/* the underlying scan relation. Same as
+								 * foreignrel, when that represents a join or
+								 * a base relation. */
 	StringInfo	buf;			/* output buffer to append to */
 	List	  **params_list;	/* exprs that will become remote Params */
 } deparse_expr_cxt;
@@ -139,9 +147,18 @@ static void mysql_append_conditions(List *exprs, deparse_expr_cxt *context);
 static void mysql_deparse_explicit_target_list(List *tlist,
 											   List **retrieved_attrs,
 											   deparse_expr_cxt *context);
-static void mysql_deparse_from_expr(StringInfo buf, PlannerInfo *root,
-									RelOptInfo *foreignrel, bool use_alias,
-									List **param_list);
+static void mysql_deparse_from_expr_for_rel(StringInfo buf, PlannerInfo *root,
+											RelOptInfo *foreignrel,
+											bool use_alias, List **param_list);
+static void mysql_deparse_from_expr(List *quals, deparse_expr_cxt *context);
+static void mysql_append_function_name(Oid funcid, deparse_expr_cxt *context);
+static void mysql_deparse_aggref(Aggref *node, deparse_expr_cxt *context);
+#if PG_VERSION_NUM >= 100000
+static void mysql_append_groupby_clause(List *tlist, deparse_expr_cxt *context);
+static Node *mysql_deparse_sort_group_clause(Index ref, List *tlist,
+											 bool force_colno,
+											 deparse_expr_cxt *context);
+#endif
 
 /*
  * Functions to construct string representation of a specific types.
@@ -241,13 +258,21 @@ mysql_deparse_select_stmt_for_rel(StringInfo buf, PlannerInfo *root,
 								  List **params_list)
 {
 	deparse_expr_cxt context;
-
-	/* We handle relations for foreign tables and joins between those */
+	List	   *quals;
 #if PG_VERSION_NUM >= 100000
-	Assert(IS_JOIN_REL(rel) || IS_SIMPLE_REL(rel));
+	MySQLFdwRelationInfo *fpinfo = (MySQLFdwRelationInfo *) rel->fdw_private;
+#endif
+
+	/*
+	 * We handle relations for foreign tables and joins between those and upper
+	 * relations.
+	 */
+#if PG_VERSION_NUM >= 100000
+	Assert(IS_JOIN_REL(rel) || IS_SIMPLE_REL(rel) || IS_UPPER_REL(rel));
 #else
 	Assert(rel->reloptkind == RELOPT_JOINREL ||
 		   rel->reloptkind == RELOPT_BASEREL ||
+		   rel->reloptkind == RELOPT_UPPER_REL ||
 		   rel->reloptkind == RELOPT_OTHER_MEMBER_REL);
 #endif
 
@@ -256,23 +281,56 @@ mysql_deparse_select_stmt_for_rel(StringInfo buf, PlannerInfo *root,
 	context.root = root;
 	context.foreignrel = rel;
 	context.params_list = params_list;
+#if PG_VERSION_NUM >= 100000
+	context.scanrel = IS_UPPER_REL(rel) ? fpinfo->outerrel : rel;
+#else
+	context.scanrel = rel;
+#endif
 
-	/* Construct SELECT clause and FROM clause */
+	/* Construct SELECT clause */
 	mysql_deparse_select_sql(tlist, retrieved_attrs, &context);
 
-	/* Construct WHERE clause */
-	if (remote_conds != NIL)
+	/*
+	 * For upper relations, the WHERE clause is built from the remote
+	 * conditions of the underlying scan relation; otherwise, we can use the
+	 * supplied list of remote conditions directly.
+	 */
+#if PG_VERSION_NUM >= 100000
+	if (IS_UPPER_REL(rel))
 	{
-		appendStringInfoString(buf, " WHERE ");
-		mysql_append_conditions(remote_conds, &context);
+		MySQLFdwRelationInfo *ofpinfo;
+
+		ofpinfo = (MySQLFdwRelationInfo *) fpinfo->outerrel->fdw_private;
+		quals = ofpinfo->remote_conds;
 	}
+	else
+#endif
+		quals = remote_conds;
+
+	/* Construct FROM and WHERE clauses */
+	mysql_deparse_from_expr(quals, &context);
+
+#if PG_VERSION_NUM >= 100000
+	if (IS_UPPER_REL(rel))
+	{
+		/* Append GROUP BY clause */
+		mysql_append_groupby_clause(fpinfo->grouped_tlist, &context);
+
+		/* Append HAVING clause */
+		if (remote_conds)
+		{
+			appendStringInfoString(buf, " HAVING ");
+			mysql_append_conditions(remote_conds, &context);
+		}
+	}
+#endif
 }
 
 /*
  * mysql_deparse_select_sql
  * 		Construct a simple SELECT statement that retrieves desired columns
  * 		of the specified foreign table, and append it to "buf".  The output
- * 		contains just "SELECT ... FROM ....".
+ * 		contains just "SELECT ...".
  *
  * tlist is the list of desired columns.  Read prologue of
  * mysql_deparse_select_stmt_for_rel() for details.
@@ -294,19 +352,17 @@ mysql_deparse_select_sql(List *tlist, List **retrieved_attrs,
 	appendStringInfoString(buf, "SELECT ");
 
 #if PG_VERSION_NUM >= 100000
-	if (IS_JOIN_REL(foreignrel))
+	if (IS_JOIN_REL(foreignrel) || IS_UPPER_REL(foreignrel))
 #else
-	if (foreignrel->reloptkind == RELOPT_JOINREL)
+	if (foreignrel->reloptkind == RELOPT_JOINREL ||
+		foreignrel->reloptkind == RELOPT_UPPER_REL)
 #endif
 	{
-		/* For a join relation use the input tlist */
-		mysql_deparse_explicit_target_list(tlist, retrieved_attrs, context);
-
 		/*
-		 * Construct FROM clause
+		 * For a join or upper relation the input tlist gives the list of
+		 * columns required to be fetched from the foreign server.
 		 */
-		appendStringInfoString(buf, " FROM ");
-		mysql_deparse_from_expr(buf, root, foreignrel, true, context->params_list);
+		mysql_deparse_explicit_target_list(tlist, retrieved_attrs, context);
 	}
 	else
 	{
@@ -327,12 +383,6 @@ mysql_deparse_select_sql(List *tlist, List **retrieved_attrs,
 		mysql_deparse_target_list(buf, root, foreignrel->relid, rel,
 								  fpinfo->attrs_used, retrieved_attrs);
 
-		/*
-		 * Construct FROM clause
-		 */
-		appendStringInfoString(buf, " FROM ");
-		mysql_deparse_relation(buf, rel);
-
 #if PG_VERSION_NUM < 130000
 		heap_close(rel, NoLock);
 #else
@@ -360,18 +410,11 @@ mysql_deparse_explicit_target_list(List *tlist, List **retrieved_attrs,
 
 	foreach(lc, tlist)
 	{
-		Var		   *var;
-
-		var = (Var *) lfirst(lc);
-		/* We expect only Var nodes here */
-		Assert(IsA(var, Var));
-
 		if (i > 0)
 			appendStringInfoString(buf, ", ");
-		mysql_deparse_var(var, context);
 
+		deparseExpr((Expr *) lfirst(lc), context);
 		*retrieved_attrs = lappend_int(*retrieved_attrs, i + 1);
-
 		i++;
 	}
 
@@ -379,6 +422,39 @@ mysql_deparse_explicit_target_list(List *tlist, List **retrieved_attrs,
 		appendStringInfoString(buf, "NULL");
 }
 
+/*
+ * mysql_deparse_from_expr
+ * 		Construct a FROM clause and, if needed, a WHERE clause, and
+ * 		append those to "buf".
+ *
+ * quals is the list of clauses to be included in the WHERE clause.
+ */
+static void
+mysql_deparse_from_expr(List *quals, deparse_expr_cxt *context)
+{
+	StringInfo	buf = context->buf;
+	RelOptInfo *scanrel = context->scanrel;
+
+	/* For upper relations, scanrel must be either a joinrel or a baserel */
+#if PG_VERSION_NUM >= 100000
+	Assert(!IS_UPPER_REL(context->foreignrel) ||
+		   IS_JOIN_REL(scanrel) || IS_SIMPLE_REL(scanrel));
+#endif
+
+	/* Construct FROM clause */
+	appendStringInfoString(buf, " FROM ");
+	mysql_deparse_from_expr_for_rel(buf, context->root, scanrel,
+									(bms_membership(scanrel->relids) == BMS_MULTIPLE),
+									context->params_list);
+
+	/* Construct WHERE clause */
+	if (quals != NIL)
+	{
+		appendStringInfoString(buf, " WHERE ");
+		mysql_append_conditions(quals, context);
+	}
+}
+
 /*
  * Deparse remote INSERT statement
  *
@@ -657,6 +733,9 @@ deparseExpr(Expr *node, deparse_expr_cxt *context)
 		case T_ArrayExpr:
 			mysql_deparse_array_expr((ArrayExpr *) node, context);
 			break;
+		case T_Aggref:
+			mysql_deparse_aggref((Aggref *) node, context);
+			break;
 		default:
 			elog(ERROR, "unsupported expression type for deparse: %d",
 				 (int) nodeTag(node));
@@ -785,8 +864,10 @@ mysql_deparse_delete(StringInfo buf, PlannerInfo *root, Index rtindex,
 static void
 mysql_deparse_var(Var *node, deparse_expr_cxt *context)
 {
-	Relids		relids = context->foreignrel->relids;
-	bool		qualify_col = (bms_membership(relids) == BMS_MULTIPLE);
+	Relids		relids = context->scanrel->relids;
+	bool		qualify_col;
+
+	qualify_col = (bms_membership(relids) == BMS_MULTIPLE);
 
 	if (bms_is_member(node->varno, relids) && node->varlevelsup == 0)
 	{
@@ -1017,26 +1098,36 @@ static void
 mysql_deparse_func_expr(FuncExpr *node, deparse_expr_cxt *context)
 {
 	StringInfo	buf = context->buf;
-	HeapTuple	proctup;
-	Form_pg_proc procform;
-	const char *proname;
 	bool		first;
 	ListCell   *arg;
 
+	/*
+	 * If the function call came from an implicit coercion, then just show the
+	 * first argument.
+	 */
+	if (node->funcformat == COERCE_IMPLICIT_CAST)
+	{
+		deparseExpr((Expr *) linitial(node->args), context);
+		return;
+	}
+
+	/* If the function call came from a cast, then show the first argument. */
+	if (node->funcformat == COERCE_EXPLICIT_CAST)
+	{
+		int32		coercedTypmod;
+
+		/* Get the typmod if this is a length-coercion function */
+		(void) exprIsLengthCoercion((Node *) node, &coercedTypmod);
+
+		deparseExpr((Expr *) linitial(node->args), context);
+		return;
+	}
+
 	/*
 	 * Normal function: display as proname(args).
 	 */
-	proctup = SearchSysCache1(PROCOID, ObjectIdGetDatum(node->funcid));
-	if (!HeapTupleIsValid(proctup))
-		elog(ERROR, "cache lookup failed for function %u", node->funcid);
-
-	procform = (Form_pg_proc) GETSTRUCT(proctup);
-
-	/* Translate PostgreSQL function into mysql function */
-	proname = mysql_replace_function(NameStr(procform->proname));
-
-	/* Deparse the function name ... */
-	appendStringInfo(buf, "%s(", proname);
+	mysql_append_function_name(node->funcid, context);
+	appendStringInfoChar(buf, '(');
 
 	/* ... and all the arguments */
 	first = true;
@@ -1049,8 +1140,6 @@ mysql_deparse_func_expr(FuncExpr *node, deparse_expr_cxt *context)
 	}
 
 	appendStringInfoChar(buf, ')');
-
-	ReleaseSysCache(proctup);
 }
 
 /*
@@ -1424,7 +1513,7 @@ foreign_expr_walker(Node *node, foreign_glob_cxt *glob_cxt,
 				 * Param's collation, i.e. it's not safe for it to have a
 				 * non-default collation.
 				 */
-				if (bms_is_member(var->varno, glob_cxt->foreignrel->relids) &&
+				if (bms_is_member(var->varno, glob_cxt->relids) &&
 					var->varlevelsup == 0)
 				{
 					/* Var belongs to foreign table */
@@ -1488,7 +1577,7 @@ foreign_expr_walker(Node *node, foreign_glob_cxt *glob_cxt,
 #endif
 
 				/* Should not be in the join clauses of the Join-pushdown */
-				if (glob_cxt->is_join_cond)
+				if (glob_cxt->is_remote_cond)
 					return false;
 
 				/* Assignment should not be in restrictions. */
@@ -1529,7 +1618,7 @@ foreign_expr_walker(Node *node, foreign_glob_cxt *glob_cxt,
 				FuncExpr   *fe = (FuncExpr *) node;
 
 				/* Should not be in the join clauses of the Join-pushdown */
-				if (glob_cxt->is_join_cond)
+				if (glob_cxt->is_remote_cond)
 					return false;
 
 				/*
@@ -1582,7 +1671,7 @@ foreign_expr_walker(Node *node, foreign_glob_cxt *glob_cxt,
 				/*
 				 * Join-pushdown allows only a few operators to be pushed down.
 				 */
-				if (glob_cxt->is_join_cond &&
+				if (glob_cxt->is_remote_cond &&
 					(!(strcmp(operatorName, "<") == 0 ||
 					   strcmp(operatorName, ">") == 0 ||
 					   strcmp(operatorName, "<=") == 0 ||
@@ -1637,7 +1726,7 @@ foreign_expr_walker(Node *node, foreign_glob_cxt *glob_cxt,
 				ScalarArrayOpExpr *oe = (ScalarArrayOpExpr *) node;
 
 				/* Should not be in the join clauses of the Join-pushdown */
-				if (glob_cxt->is_join_cond)
+				if (glob_cxt->is_remote_cond)
 					return false;
 
 				/*
@@ -1730,7 +1819,7 @@ foreign_expr_walker(Node *node, foreign_glob_cxt *glob_cxt,
 				ArrayExpr  *a = (ArrayExpr *) node;
 
 				/* Should not be in the join clauses of the Join-pushdown */
-				if (glob_cxt->is_join_cond)
+				if (glob_cxt->is_remote_cond)
 					return false;
 
 				/*
@@ -1780,6 +1869,91 @@ foreign_expr_walker(Node *node, foreign_glob_cxt *glob_cxt,
 				check_type = false;
 			}
 			break;
+		case T_Aggref:
+			{
+				Aggref	   *agg = (Aggref *) node;
+				ListCell   *lc;
+				const char *func_name = get_func_name(agg->aggfnoid);
+
+#if PG_VERSION_NUM >= 100000
+				/* Not safe to pushdown when not in grouping context */
+				if (!IS_UPPER_REL(glob_cxt->foreignrel))
+					return false;
+
+				/* Only non-split aggregates are pushable. */
+				if (agg->aggsplit != AGGSPLIT_SIMPLE)
+					return false;
+#endif
+
+				/* As usual, it must be shippable. */
+				if (!is_builtin(agg->aggfnoid))
+					return false;
+
+				if (!(strcmp(func_name, "min") == 0 ||
+					strcmp(func_name, "max") == 0 ||
+					strcmp(func_name, "sum") == 0 ||
+					strcmp(func_name, "avg") == 0 ||
+					strcmp(func_name, "count") == 0))
+					return false;
+
+
+				/*
+				 * Recurse to input args. aggdirectargs, aggorder and
+				 * aggdistinct are all present in args, so no need to check
+				 * their shippability explicitly.
+				 */
+				foreach(lc, agg->args)
+				{
+					Node	   *n = (Node *) lfirst(lc);
+
+					/* If TargetEntry, extract the expression from it */
+					if (IsA(n, TargetEntry))
+					{
+						TargetEntry *tle = (TargetEntry *) n;
+
+						n = (Node *) tle->expr;
+					}
+
+					if (!foreign_expr_walker(n, glob_cxt, &inner_cxt))
+						return false;
+				}
+
+				/* Aggregates with order are not supported on MySQL. */
+				if (agg->aggorder)
+					return false;
+
+				/* FILTER clause is not supported on MySQL. */
+				if (agg->aggfilter)
+					return false;
+
+				/*
+				 * If aggregate's input collation is not derived from a
+				 * foreign Var, it can't be sent to remote.
+				 */
+				if (agg->inputcollid == InvalidOid)
+					 /* OK, inputs are all noncollatable */ ;
+				else if (inner_cxt.state != FDW_COLLATE_SAFE ||
+						 agg->inputcollid != inner_cxt.collation)
+					return false;
+
+				/*
+				 * Detect whether node is introducing a collation not derived
+				 * from a foreign Var.  (If so, we just mark it unsafe for now
+				 * rather than immediately returning false, since the parent
+				 * node might not care.)
+				 */
+				collation = agg->aggcollid;
+				if (collation == InvalidOid)
+					state = FDW_COLLATE_NONE;
+				else if (inner_cxt.state == FDW_COLLATE_SAFE &&
+						 collation == inner_cxt.collation)
+					state = FDW_COLLATE_SAFE;
+				else if (collation == DEFAULT_COLLATION_OID)
+					state = FDW_COLLATE_NONE;
+				else
+					state = FDW_COLLATE_UNSAFE;
+			}
+			break;
 		default:
 
 			/*
@@ -1850,10 +2024,13 @@ foreign_expr_walker(Node *node, foreign_glob_cxt *glob_cxt,
  */
 bool
 mysql_is_foreign_expr(PlannerInfo *root, RelOptInfo *baserel, Expr *expr,
-					  bool is_join_cond)
+					  bool is_remote_cond)
 {
 	foreign_glob_cxt glob_cxt;
 	foreign_loc_cxt loc_cxt;
+#if PG_VERSION_NUM >= 100000
+	MySQLFdwRelationInfo *fpinfo = (MySQLFdwRelationInfo *) (baserel->fdw_private);
+#endif
 
 	/*
 	 * Check that the expression consists of nodes that are safe to execute
@@ -1861,15 +2038,31 @@ mysql_is_foreign_expr(PlannerInfo *root, RelOptInfo *baserel, Expr *expr,
 	 */
 	glob_cxt.root = root;
 	glob_cxt.foreignrel = baserel;
-	glob_cxt.is_join_cond = is_join_cond;
+	glob_cxt.is_remote_cond = is_remote_cond;
+
+	/*
+	 * For an upper relation, use relids from its underneath scan relation,
+	 * because the upperrel's own relids currently aren't set to anything
+	 * meaningful by the core code.  For other relation, use their own relids.
+	 */
+#if PG_VERSION_NUM >= 100000
+	if (IS_UPPER_REL(baserel))
+		glob_cxt.relids = fpinfo->outerrel->relids;
+	else
+#endif
+		glob_cxt.relids = baserel->relids;
+
 	loc_cxt.collation = InvalidOid;
 	loc_cxt.state = FDW_COLLATE_NONE;
 	if (!foreign_expr_walker((Node *) expr, &glob_cxt, &loc_cxt))
 		return false;
 
-	/* Expressions examined here should be boolean, ie noncollatable */
-	Assert(loc_cxt.collation == InvalidOid);
-	Assert(loc_cxt.state == FDW_COLLATE_NONE);
+	/*
+	 * If the expression has a valid collation that does not arise from a
+	 * foreign var, the expression can not be sent over.
+	 */
+	if (loc_cxt.state == FDW_COLLATE_UNSAFE)
+		return false;
 
 	/*
 	 * An expression which includes any mutable functions can't be sent over
@@ -1889,7 +2082,8 @@ mysql_is_foreign_expr(PlannerInfo *root, RelOptInfo *baserel, Expr *expr,
  * mysql_append_conditions
  * 		Deparse conditions from the provided list and append them to buf.
  *
- * The conditions in the list are assumed to be ANDed.
+ * The conditions in the list are assumed to be ANDed.  This function is used
+ * to deparse WHERE clauses, JOIN .. ON clauses and HAVING clauses.
  *
  * Depending on the caller, the list elements might be either RestrictInfos
  * or bare clauses.
@@ -1929,13 +2123,13 @@ mysql_append_conditions(List *exprs, deparse_expr_cxt *context)
 }
 
 /*
- * mysql_deparse_from_expr
+ * mysql_deparse_from_expr_for_rel
  * 		Construct a FROM clause
  */
 void
-mysql_deparse_from_expr(StringInfo buf, PlannerInfo *root,
-						RelOptInfo *foreignrel, bool use_alias,
-						List **params_list)
+mysql_deparse_from_expr_for_rel(StringInfo buf, PlannerInfo *root,
+								RelOptInfo *foreignrel, bool use_alias,
+								List **params_list)
 {
 	MySQLFdwRelationInfo *fpinfo = (MySQLFdwRelationInfo *) foreignrel->fdw_private;
 
@@ -1952,11 +2146,13 @@ mysql_deparse_from_expr(StringInfo buf, PlannerInfo *root,
 
 		/* Deparse outer relation */
 		initStringInfo(&join_sql_o);
-		mysql_deparse_from_expr(&join_sql_o, root, rel_o, true, params_list);
+		mysql_deparse_from_expr_for_rel(&join_sql_o, root, rel_o, true,
+										params_list);
 
 		/* Deparse inner relation */
 		initStringInfo(&join_sql_i);
-		mysql_deparse_from_expr(&join_sql_i, root, rel_i, true, params_list);
+		mysql_deparse_from_expr_for_rel(&join_sql_i, root, rel_i, true,
+										params_list);
 
 		/*
 		 * For a join relation FROM clause entry is deparsed as
@@ -1974,6 +2170,7 @@ mysql_deparse_from_expr(StringInfo buf, PlannerInfo *root,
 
 			context.buf = buf;
 			context.foreignrel = foreignrel;
+			context.scanrel = foreignrel;
 			context.root = root;
 			context.params_list = params_list;
 
@@ -2048,3 +2245,224 @@ mysql_get_jointype_name(JoinType jointype)
 	/* Keep compiler happy */
 	return NULL;
 }
+
+/*
+ * mysql_append_function_name
+ *		Deparses function name from given function oid.
+ */
+static void
+mysql_append_function_name(Oid funcid, deparse_expr_cxt *context)
+{
+	StringInfo	buf = context->buf;
+	HeapTuple	proctup;
+	Form_pg_proc procform;
+	const char *proname;
+
+	proctup = SearchSysCache1(PROCOID, ObjectIdGetDatum(funcid));
+	if (!HeapTupleIsValid(proctup))
+		elog(ERROR, "cache lookup failed for function %u", funcid);
+	procform = (Form_pg_proc) GETSTRUCT(proctup);
+
+	/* Always print the function name */
+	proname = NameStr(procform->proname);
+
+	/* Translate PostgreSQL function into MySQL function */
+	proname = mysql_replace_function(NameStr(procform->proname));
+
+	appendStringInfoString(buf, proname);
+
+	ReleaseSysCache(proctup);
+}
+
+/*
+ * mysql_deparse_aggref
+ *		Deparse an Aggref node.
+ */
+static void
+mysql_deparse_aggref(Aggref *node, deparse_expr_cxt *context)
+{
+	StringInfo	buf = context->buf;
+	bool		use_variadic;
+
+	/* Only basic, non-split aggregation accepted. */
+	Assert(node->aggsplit == AGGSPLIT_SIMPLE);
+
+	/* Check if need to print VARIADIC (cf. ruleutils.c) */
+	use_variadic = node->aggvariadic;
+
+	/* Find aggregate name from aggfnoid which is a pg_proc entry. */
+	mysql_append_function_name(node->aggfnoid, context);
+	appendStringInfoChar(buf, '(');
+
+	/* Add DISTINCT */
+	appendStringInfoString(buf, (node->aggdistinct != NIL) ? "DISTINCT " : "");
+
+	/* aggstar can be set only in zero-argument aggregates. */
+	if (node->aggstar)
+		appendStringInfoChar(buf, '*');
+	else
+	{
+		ListCell   *arg;
+		bool		first = true;
+
+		/* Add all the arguments. */
+		foreach(arg, node->args)
+		{
+			TargetEntry *tle = (TargetEntry *) lfirst(arg);
+			Node	   *n = (Node *) tle->expr;
+
+			if (tle->resjunk)
+				continue;
+
+			if (!first)
+				appendStringInfoString(buf, ", ");
+			first = false;
+
+			/* Add VARIADIC */
+#if PG_VERSION_NUM >= 130000
+			if (use_variadic && lnext(node->args, arg) == NULL)
+#else
+			if (use_variadic && lnext(arg) == NULL)
+#endif
+				appendStringInfoString(buf, "VARIADIC ");
+
+			deparseExpr((Expr *) n, context);
+		}
+	}
+
+	appendStringInfoChar(buf, ')');
+}
+
+/*
+ * mysql_append_groupby_clause
+ * 		Deparse GROUP BY clause.
+ */
+#if PG_VERSION_NUM >= 100000
+static void
+mysql_append_groupby_clause(List *tlist, deparse_expr_cxt *context)
+{
+	StringInfo	buf = context->buf;
+	Query	   *query = context->root->parse;
+	ListCell   *lc;
+	bool		first = true;
+
+	/* Nothing to be done, if there's no GROUP BY clause in the query. */
+	if (!query->groupClause)
+		return;
+
+	appendStringInfoString(buf, " GROUP BY ");
+
+	/*
+	 * Queries with grouping sets are not pushed down, so we don't expect
+	 * grouping sets here.
+	 */
+	Assert(!query->groupingSets);
+
+	foreach(lc, query->groupClause)
+	{
+		SortGroupClause *grp = (SortGroupClause *) lfirst(lc);
+
+		if (!first)
+			appendStringInfoString(buf, ", ");
+		first = false;
+
+		mysql_deparse_sort_group_clause(grp->tleSortGroupRef, tlist,
+										true, context);
+	}
+}
+
+/*
+ * mysql_deparse_sort_group_clause
+ * 		Appends a sort or group clause.
+ */
+static Node *
+mysql_deparse_sort_group_clause(Index ref, List *tlist, bool force_colno,
+								deparse_expr_cxt *context)
+{
+	StringInfo	buf = context->buf;
+	TargetEntry *tle;
+	Expr	   *expr;
+
+	tle = get_sortgroupref_tle(ref, tlist);
+	expr = tle->expr;
+
+	if (force_colno)
+	{
+		/* Use column-number form when requested by caller. */
+		Assert(!tle->resjunk);
+		appendStringInfo(buf, "%d", tle->resno);
+	}
+	else if (expr && IsA(expr, Const))
+	{
+		/*
+		 * Force a typecast here so that we don't emit something like "GROUP
+		 * BY 2", which will be misconstrued as a column position rather than
+		 * a constant.
+		 */
+		mysql_deparse_const((Const *) expr, context);
+	}
+	else if (!expr || IsA(expr, Var))
+		deparseExpr(expr, context);
+	else
+	{
+		/* Always parenthesize the expression. */
+		appendStringInfoChar(buf, '(');
+		deparseExpr(expr, context);
+		appendStringInfoChar(buf, ')');
+	}
+
+	return (Node *) expr;
+}
+#endif
+
+/*
+ * mysql_is_foreign_param
+ * 		Returns true if given expr is something we'd have to send the
+ * 		value of to the foreign server.
+ *
+ * This should return true when the expression is a shippable node that
+ * deparseExpr would add to context->params_list.  Note that we don't care
+ * if the expression *contains* such a node, only whether one appears at top
+ * level.  We need this to detect cases where setrefs.c would recognize a
+ * false match between an fdw_exprs item (which came from the params_list)
+ * and an entry in fdw_scan_tlist (which we're considering putting the given
+ * expression into).
+ */
+bool
+mysql_is_foreign_param(PlannerInfo *root,
+					   RelOptInfo *baserel,
+					   Expr *expr)
+{
+	if (expr == NULL)
+		return false;
+
+	switch (nodeTag(expr))
+	{
+		case T_Var:
+			{
+				/* It would have to be sent unless it's a foreign Var. */
+				Var		   *var = (Var *) expr;
+				Relids		relids;
+#if PG_VERSION_NUM >= 100000
+				MySQLFdwRelationInfo *fpinfo = (MySQLFdwRelationInfo *) (baserel->fdw_private);
+
+				if (IS_UPPER_REL(baserel))
+					relids = fpinfo->outerrel->relids;
+				else
+#endif
+					relids = baserel->relids;
+
+				if (bms_is_member(var->varno, relids) && var->varlevelsup == 0)
+					return false;	/* foreign Var, so not a param. */
+				else
+					return true;	/* it'd have to be a param. */
+				break;
+			}
+		case T_Param:
+			/* Params always have to be sent to the foreign server. */
+			return true;
+		default:
+			break;
+	}
+	return false;
+}
