@@ -1,13 +1,13 @@
 /*-------------------------------------------------------------------------
  *
- * mysql_fdw--1.0.sql
+ * mysql_fdw--1.1.sql
  * 			Foreign-data wrapper for remote MySQL servers
  *
  * Portions Copyright (c) 2012-2014, PostgreSQL Global Development Group
  * Portions Copyright (c) 2004-2021, EnterpriseDB Corporation.
  *
  * IDENTIFICATION
- * 			mysql_fdw--1.0.sql
+ * 			mysql_fdw--1.1.sql
  *
  *-------------------------------------------------------------------------
  */
@@ -26,3 +26,7 @@ LANGUAGE C STRICT;
 CREATE FOREIGN DATA WRAPPER mysql_fdw
   HANDLER mysql_fdw_handler
   VALIDATOR mysql_fdw_validator;
+
+CREATE OR REPLACE FUNCTION mysql_fdw_version()
+  RETURNS pg_catalog.int4 STRICT
+  AS 'MODULE_PATHNAME' LANGUAGE C;
