@@ -300,9 +300,8 @@ SELECT * FROM f_enum_t1 WHERE size = 'medium' ORDER BY id;
 
 -- Remote aggregate in combination with a local Param (for the output
 -- of an initplan)
-EXPLAIN (VERBOSE, COSTS OFF)
-SELECT EXISTS(SELECT 1 FROM pg_enum), sum(id) from f_enum_t1;
 SELECT EXISTS(SELECT 1 FROM pg_enum), sum(id) from f_enum_t1;
+SELECT EXISTS(SELECT 1 FROM pg_enum), sum(id) from f_enum_t1 GROUP BY 1;
 
 -- Check with the IMPORT FOREIGN SCHEMA command.  Also, check ENUM types with
 -- the IMPORT FOREIGN SCHEMA command. If the enum name is the same for multiple
@@ -363,10 +362,6 @@ SELECT c1, c2 FROM f_test_tbl2 WHERE c3 = (SELECT 'PUNE'::varchar) ORDER BY c1;
 SELECT * FROM local_t1 lt1 WHERE lt1.c1 =
   (SELECT count(*) FROM f_test_tbl3 ft1 WHERE ft1.c2 = lt1.c2) ORDER BY lt1.c1;
 
-EXPLAIN (VERBOSE, COSTS OFF)
-SELECT c1, c2 FROM f_test_tbl1 WHERE c8 = (
-  SELECT c1 FROM f_test_tbl2 WHERE c1 = (
-    SELECT min(c1) + 1 FROM f_test_tbl2)) ORDER BY c1;
 SELECT c1, c2 FROM f_test_tbl1 WHERE c8 = (
   SELECT c1 FROM f_test_tbl2 WHERE c1 = (
     SELECT min(c1) + 1 FROM f_test_tbl2)) ORDER BY c1;
@@ -419,6 +414,19 @@ SELECT c1, c2, c3 FROM f_test_tbl1 WHERE pg_catalog.timeofday() IS NOT NULL
 SELECT c1, c2, c3 FROM f_test_tbl1 WHERE pg_catalog.timeofday() IS NOT NULL
   ORDER BY 1 limit 5;
 
+-- FDW-447: Fix function implicit/explicit coercion.
+EXPLAIN (VERBOSE, COSTS OFF)
+SELECT * FROM f_test_tbl1 WHERE c1 = 12.2;
+EXPLAIN (VERBOSE, COSTS OFF)
+SELECT * FROM f_test_tbl1 WHERE c1::numeric = 12.2;
+
+-- FDW-408: Skip importing relations that have SET type because Postgres
+-- doesn't have equivalent datatype which can be mapped to MySQL SET.
+IMPORT FOREIGN SCHEMA mysql_fdw_regress LIMIT TO (test_set, mysql_test, test_tbl1)
+  FROM SERVER mysql_svr INTO public;
+SELECT relname FROM pg_class
+  WHERE relname IN ('test_set', 'mysql_test', 'test_tbl1') AND relnamespace = 'public'::regnamespace;
+
 -- Cleanup
 DROP TABLE l_test_tbl1;
 DROP TABLE l_test_tbl2;
@@ -440,6 +448,8 @@ DROP FOREIGN TABLE f_test_tbl3;
 DROP FOREIGN TABLE test5;
 DROP FOREIGN TABLE test5_1;
 DROP FOREIGN TABLE test5_2;
+DROP FOREIGN TABLE mysql_test;
+DROP FOREIGN TABLE test_tbl1;
 DROP TYPE size_t;
 DROP TYPE enum_t1_size_t;
 DROP TYPE enum_t2_size_t;
